--[[
	/####################################\
	##                                  ##
	##            SlotNames             ##
	##      Einfachere Verwendung       ##
	##       von benannten Slots        ##
	##                                  ##
	##   Autor: Benjamin Hogl (BH2),    ##
	##          benjamin.hogl@gmx.de    ##
	##   Version 0.1.1, 2018-12-27      ##
	##                                  ##
	\####################################/
]]

local slotNrMin, slotNrMax = 1, 1000

-- stores the mapping from names to slotnumbers. 
-- Can be modified outside by the user; can be overwritten by calling SlotFuncs.setMapping(...)
local SlotMapping = {}

-- table holding all functions callable by the user (some may be used internally, too)
local SlotFuncs = {}

function SlotFuncs.getVersion()
	return {0,1,1}
end

-- looks up a slot number in the mapping table, given the path as a list of argument
function SlotFuncs.lookupSlotNr(...)
	local function lookupSlotNrHelper(mappingTable, firstKey, ...)
		if firstKey==nil then 
			return mappingTable
		elseif type(mappingTable) == "table" then
			return lookupSlotNrHelper(mappingTable[firstKey], ...)
		else
			return nil
		end
	end
	return lookupSlotNrHelper(SlotMapping, ...)
end

-- returns a string representing the path given as a table
function SlotFuncs.getPathString(path)
	return table.concat(path, ".")
end

-- adds a mapping for all the slotnumbers, so you can access slot 1 via Slot[1]
-- if a number is already mapped to some slot, it won't get overwritten
function SlotFuncs.addDefaultMapping()
	for i=slotNrMin,slotNrMax do
		if not SlotMapping[i] then -- don't overwrite existing fields
			SlotMapping[i] = i
		end
	end
end

-- removes the default mapping for the slotnumbers. Numbers with other mappings aren't touched
function SlotFuncs.removeDefaultMapping()
	for i=slotNrMin,slotNrMax do
		if SlotMapping[i] == i then -- only remove default values
			SlotMapping[i] = nil
		end
	end
end

-- replaces the current mapping table with a new ohne
function SlotFuncs.setMapping(newMapping)
	SlotMapping = newMapping
end

-- Checks the mapping table for possible duplicates (multiple paths mapped to the same slot number),
-- slot numbers that are not numbers or not in the allowed range
-- A table can be passed containing the following options: 
--[[
{
	checkNumber=true				-- check if each slotNr is really a number
	checkRange=true					-- check if each slotNr is in the range from 1..1000
	checkDuplicates=true			-- check if a slotNr is used multiple times (from different paths)
	results="error" 				-- or "print" or "return": what to do with the found errors
	ignoreDefaultMapping=false		-- or true: whether the mapping of numbers to themselves should be considered as duplicates
	printSuccess=false				-- or true: whether a message should be printed even if there is no error
}
]]--
function SlotFuncs.checkMapping(options)
	if type(options) ~= "table" then options = {} end
	setmetatable(options, {__index={checkNumber=true, checkRange=true, checkDuplicates=true, ignoreDefaultMapping=false, printSuccess=false, results="error"}})
	
	local mappingSlotNrToPath = {}
	local foundErrors = { numChecked=0 }
		
	-- checks whether this is a valid slotNr. Returns nil if valid, an error message otherwise.
	local function checkSlotNr(slotNr)
		if options.checkNumber and type(slotNr) ~= "number" then
			return string.format("%q vom Typ %s ist keine gltige Slot-Nummer", slotNr, type(slotNr))
		elseif options.checkRange and slotNr < slotNrMin or slotNr > slotNrMax then
			return string.format("%s ist keine gltige Slot-Nummer, da sie auerhalb des Bereichs von %i bis %i liegt", slotNr, slotNrMin, slotNrMax)
		elseif options.checkDuplicates and mappingSlotNrToPath[slotNr] then
			return string.format("Slot %s wird schon von %q verwendet", slotNr, SlotFuncs.getPathString(mappingSlotNrToPath[slotNr]))
		else
			return nil
		end
	end
	
	local function checkMappingHelper(mappingTable, ...)
		local currentPath = {...}
		if type(mappingTable) == "table" then 
			-- check this table recursively
			currentPath[#currentPath+1] = "" -- reserve space for setting the last key later
			for key, value in pairs(mappingTable) do
				currentPath[#currentPath] = key
				checkMappingHelper(value, table.unpack(currentPath))
			end
		else
			if options.ignoreDefaultMapping and #currentPath == 1 and mappingTable == currentPath[1] then return end
			local checkResult = checkSlotNr(mappingTable) -- mappingTable is a slotNr
			if checkResult then
				table.insert(foundErrors, { path=currentPath, message=checkResult })
			else
				mappingSlotNrToPath[mappingTable] = currentPath
			end
			foundErrors.numChecked = foundErrors.numChecked+1
		end
	end
	checkMappingHelper(SlotMapping)
	
	if options.results == "return" then return foundErrors
	elseif #foundErrors > 0 then
		print()
		for _,foundError in ipairs(foundErrors) do
			print(string.format("%s: %s", SlotFuncs.getPathString(foundError.path), foundError.message))
		end
		print("------------------------------------------------------------------------------------------------------------------------")
		local message = string.format("Bei der berprfung von %i Slot-Mapping-Eintrgen wurden %i Fehler gefunden!", foundErrors.numChecked, #foundErrors)
		if options.results == "print" then 
			print(message)
		else
			error(message, 0)
		end
	elseif options.printSuccess then
		print(string.format("Bei der berprfung von %i Slot-Mapping-Eintrgen wurden keine Fehler gefunden!", foundErrors.numChecked))
	end
	return (#foundErrors <= 0)
end

-- creates a table with metamethods for reading and setting to slots, initialized with the given path
local function getSlotTable(...)
	local path = {...}
	local currentPath = {...}
	
	local Slot = {}
	local mt = {}
	function mt:__index(key) -- wenn `Slot[key]` abgerufen wird
		if key == nil then return nil end
		currentPath[#path+1] = key;
		local nr = SlotFuncs.lookupSlotNr(table.unpack(currentPath))
		if type(nr) == "table" then  -- to enable something like Slot.Bhf.Gleis1, return the table for Slot.Bhf
			return getSlotTable(table.unpack(currentPath))
		end
		if type(nr) ~= "number" then -- avoid error message from EEP, print our own instead
			error(string.format("\"%s\" ist keine gltige Slotnummer zugeordnet (sondern %s), und kann deshalb nicht ausgelesen werden!", SlotFuncs.getPathString(currentPath), nr), 2)
			return nil
		end
		local erfolg, wert = EEPLoadData(nr)
		if erfolg then
			return wert
		else
			return nil
		end
	end

	function mt:__newindex(key, wert) -- wenn `Slot[nr] = wert` zugewiesen wird
		if key == nil then error("table index is nil",2); return end
		currentPath[#path+1] = key;
		local nr = SlotFuncs.lookupSlotNr(table.unpack(currentPath))
		if type(nr) ~= "number" then  -- avoid error message from EEP, print our own instead
			error(string.format("\"%s\" ist keine gltige Slotnummer zugeordnet (sondern %s), deshalb kann der Wert %s nicht gespeichert werden!", SlotFuncs.getPathString(currentPath), nr, wert), 2)
			return
		end
		local erfolg = EEPSaveData(nr, wert)
		if not erfolg then
			error(string.format("Der Wert %s konnte nicht in Slot %i (\"%s\") gespeichert werden!", wert, nr, SlotFuncs.getPathString(currentPath)), 2)
		end
	end

	-- enable length operator, ipairs and pairs for this table. The keys are taken from the corresponding mapping table
	function mt:__len()
		return #SlotFuncs.lookupSlotNr(table.unpack(path))
	end
	function mt:__ipairs()
		local i, n = 0, #self
		return function() i = i + 1; if i<=n then return i, self[i] end end
	end
	function mt:__pairs()
		local mappingTable = SlotFuncs.lookupSlotNr(table.unpack(path))
		return function(_,k) k = next(mappingTable, k); if k~=nil then return k, self[k] end end
	end

	setmetatable(Slot, mt)
	return Slot
end

print ("SlotNames_BH2 v"..table.concat(SlotFuncs.getVersion(),'.').." wurde eingebunden")

-- require(...) can only return one value. This value is a function returning three values, so you can write
--		Slot, SlotMapping, SlotSugar = require("SlotNames_BH2")()
local rootSlotTable = getSlotTable()
return function()
	return rootSlotTable, SlotMapping, SlotFuncs
end
